<'
# Copyright (c) 2011-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Fri Mar  2 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_c_types.h"
#include "typecopy.h"
#include "<"$comp">_parameters.h"
#include "<"$comp">_control_task.h"


/* --- genom_connect_port_codel -------------------------------------------- */

genom_event
genom_connect_port_codel(const char local[128], const char remote[128])
{
  /* find the inport */
<'foreach p [$component ports in] {'>
<'  if {"multiple" in [$p kind]} {'>
  if (!strncmp(local, "<"[$p name]">/", sizeof("<"[$p name]">/")-1)) {
    genom_event s = genom_<"$comp">_<"[$p name]">_connect(
      local + sizeof("<"[$p name]">/")-1, remote, genom_<"$comp">_cids());
    if (s) return s;
  }
<'  } else {'>
  if (!strcmp(local, "<"[$p name]">")) {
    genom_event s;
    s = genom_<"$comp">_<"[$p name]">_connect(remote, genom_<"$comp">_cids());
    if (s) return s;
  }
<'  }'>
<'}'>

  return genom_ok;
}


/* --- genom_connect_remote_codel ------------------------------------------ */

genom_event
genom_connect_remote_codel(const char local[128], const char remote[128])
{
  /* find the remote */
<'foreach r [$component remotes] {'>
  if (!strcmp(local, "<"[$r name]">")) {
    genom_event s;
    s = genom_<"$comp">_<"[$r name]">_connect(remote, genom_<"$comp">_cids());
    if (s) return s;
  }
<'}'>

  return genom_ok;
}


<'foreach s [$component services] {'>

/* <"[--- Service [$s name] control codels ------------------------------]"> */

genom_event
genom_<"$comp">_<"[$s name]">_controlcb(
  struct genom_<"$comp">_cntrlstr *cids,
  struct genom_<"$comp">_<"[$s name]">_activity *a)
{
  genom_event s = genom_ok;

  /* check allowance (before/after statements) */
<'foreach other [$s after] {'>
  if (!cids->run_map[<"$COMP">_<"[$other name]">_RQSTID])
    return genom_disallowed();
<'}'>
<'foreach other [$s before] {'>
  if (cids->run_map[<"$COMP">_<"[$other name]">_RQSTID])
    return genom_disallowed();
<'}'>

  /* copy inout parameters to output */
<'  foreach p [$s parameters inout] {'>
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->out.[$p name]]">,
    <"[[$p type] pass value a->in.[$p name]]">);
<'  }'>

  /* call validate codel */
<'  if {[llength [$s validate]]} {'>
<'    # build parameter list
    set plist [list]
    foreach p [[$s validate] parameters] {
      switch -- [$p src] {
        ids	{ set m "cids->ids[$p member]" }
        local	{
          switch -- [[$p param] dir] {
            local	{ set m "a->locals.[$p name]" }
            in		{ set m "a->in.[$p name]" }
            inout - out { set m "a->out.[$p name]" }
            default	{ error "unsupported local direction" }
          }
        }
        default	{ error "unsupported validate codel parameter" }
      }
      switch -- [$p dir] {
	in		{ lappend plist [[$p type] pass value $m] }
	inout - out	{ lappend plist [[$p type] pass reference $m] }
        default		{ error "unsupported direction" }
      }
    }'>
  s = <"[[$s validate] invoke $plist]">;
  genom_log_debug("service <"[$s name]"> validation returned %s", s?s:"ok");
  if (s) return s;
<'  }'>

  /* copy ids parameters to ids (attributes) */
<'  foreach p [$s parameters in inout] {'>
<'    if {[$p src] == "ids"} {'>
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference cids->ids[$p member]]">,
    <"[[$p type] pass value a->in.[$p name]]">);
<'    }'>
<'  }'>
<'  foreach p [$s parameters out inout] {'>
<'    if {[$p src] == "ids"} {'>
  genom_tcopy_<"[[$p type] mangle]">(
    <"[[$p type] pass reference a->out.[$p name]]">,
    <"[[$p type] pass value cids->ids[$p member]]">);
<'    }'>
<'  }'>

  /* call simple codels */
<'foreach c [$s codels simple] {'>
<'  # build parameter list
    set plist [list]
    foreach p [$c parameters] {
      switch -- [$p src] {
        ids	{ set m "cids->ids[$p member]" }
        local	{
          switch -- [[$p param] dir] {
            local	{ set m "a->locals.[$p name]" }
            in		{ set m "a->in.[$p name]" }
            inout - out { set m "a->out.[$p name]" }
            default	{ error "unsupported local direction" }
          }
        }
        default		{ error "unsupported codel parameter" }
      }
      switch -- [$p dir] {
	in		{ lappend plist [[$p type] pass value $m] }
	inout - out	{ lappend plist [[$p type] pass reference $m] }
        default		{ error "unsupported direction" }
      }
    }'>
  s = <"[$c invoke $plist]">;
  genom_log_debug("service %s codel %s returned %s",
                  "<"[$s name]">", "<"[$c name]">", s?s:"ok");
  if (s
<'    foreach t [$s throw] {'>
      && s != <"[$t cname]">_id
<'    }'>
      ) {
    genom_unkex_detail d;
    strncpy(d.what, s, sizeof(d.what)); d.what[sizeof(d.what)-1] = *"";
    genom_log_warn(
      0, "unknown exception %s for codel %s in service %s",
      s, "<"[$c name]">", "<"[$s name]">");
    s = genom_unkex(&d);
  }
  if (s) return s;
<'  }'>

  return s;
}

<'}'>
