<'
# Copyright (c) 2011-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Mon Mar  5 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "commonStructLib.h"
#include "h2evnLib.h"
<'if {![catch {$task period}]} {'>
#include "h2timerLib.h"
<'}'>
#include "taskLib.h"

#include "<"$comp">_control_task.h"


/* --- local data ---------------------------------------------------------- */

/* maximum <"$tname"> mbox size */
static __inline__ size_t
<"$comp">_<"$tname">_mbox_size(void)
{
<'  set n 0 '>
<'  foreach c [$task codels] {'>
<'    foreach p [$c parameters remote] {'>
<'      if {[incr n] == 1} {'>
  size_t s, m = 0;
<'      }'>
  s = genom_<"$comp">_<"[[$p remote] name]">_maxreply_size();
  if (m < s) m = s;
<'    }'>
<'  }'>
<'  foreach s [$task services] {'>
<'    foreach c [$s codels] {'>
<'      foreach p [$c parameters remote] {'>
<'      if {[incr n] == 1} {'>
  size_t s, m = 0;
<'      }'>
  s = genom_<"$comp">_<"[[$p remote] name]">_maxreply_size();
  if (m < s) m = s;
<'      }'>
<'    }'>
<'  }'>
<'  if {$n} {'>
  return (MBOX_BUF_SIZE(m)+MBOX_BUF_SIZE(sizeof(int)))*CLIENT_NMAX_RQST_ID;
<'  } else {'>
  return 0;
<'  }'>
}


/* <"[--- genom_${comp}_${tname}_exec_task ------------------------------]"> */

void *
genom_<"$comp">_<"$tname">_exec_task(void *data)
{
  struct genom_<"$comp">_cntrlstr *cids = data;
  struct genom_activity *a;
<'if {![catch {$task period}]} {'>
  H2TIMER_ID timer;
<'}'>
  int sched;
  int id;

<'if {[catch {$task period} period]} {'>
  /* call mboxInit: h2evnSusp() relies on it and we want to catch errors during
   * startup */
  if (mboxInit(NULL) == ERROR) {
    genom_log_warn(1, "cannot create %s exec task", "<"$tname">");
    semGive(cids->tasks.<"$tname">.sync);
    return NULL;
  }
<'} else {'>
<'
  set ticks [expr {[$period value]/[[$component clock-rate] value]}]
  if {abs($ticks - round($ticks)) > 1e-6} {
    error [concat \
               "task $tname period ([$period value]) not a multiple of" \
               "clock-rate ([[$component clock-rate] value])"]
  }
  set ticks [expr round($ticks)]

  if {[catch {[$task delay] value} delay]} { set delay 0 }
  set delay [expr {$delay/[[$component clock-rate] value]}]
  if {abs($delay - round($delay)) > 1e-6} {
    error [concat \
               "task $tname delay ($delay) not a multiple of" \
               "clock-rate ([[$component clock-rate] value])"]
  }
  set delay [expr round($delay)]
'>
  /* create an h2 timer */
  timer = h2timerAlloc();
  if (timer == NULL) {
    genom_log_warn(1, "cannot create timer for %s exec task", "<"$tname">");
    semGive(cids->tasks.<"$tname">.sync);
    return NULL;
  }

  if (h2timerStart(timer, <"$ticks">, <"$delay">) != OK) {
    genom_log_warn(1, "cannot start timer for %s exec task", "<"$tname">");
    semGive(cids->tasks.<"$tname">.sync);
    return NULL;
  }
<'}'>

  /* initialize csLib for remotes */
  if (<"$comp">_<"$tname">_mbox_size()) {
    char mbox[64];
    snprintf(mbox, sizeof(mbox), "%s/%s", genom_instance, "<"$tname">");

    if (csMboxInit(mbox, 0, <"$comp">_<"$tname">_mbox_size()) == ERROR) {
      genom_log_warn(1, "cannot create %s task mailbox %s",
                     "<"$tname">", mbox);
<'if {[catch {$task period} period]} {'>
      mboxEnd(0);
<'}'>
      semGive(cids->tasks.<"$tname">.sync);
      return NULL;
    }
  }

  /* init status */
  commonStructTake(cids);
  sched = 0;
  cids->tasks.<"$tname">.status = genom_ok;
  cids->tasks.<"$tname">.pact = <"[[dotgen type ::${comp}::start] cname]">;

  /* permanent activity start codel */
  sched = genom_<"$comp">_<"$tname">_invoke_perm(cids);
  if (sched < 0) {
    commonStructGive(cids);
    if (<"$comp">_<"$tname">_mbox_size()) {
      csMboxEnd();
    }
<'if {[catch {$task period} period]} {'>
    mboxEnd(0);
<'} else {'>
    h2timerStop(timer);
    h2timerFree(timer);
<'}'>
    cids->tasks.<"$tname">.status = cids->tasks.<"$tname">.pact;
    semGive(cids->tasks.<"$tname">.sync);
    return NULL;
  }

  /* signal parent */
  genom_log_info("spawned %s exec task", "<"$tname">");
  cids->tasks.<"$tname">.taskid = taskIdSelf();
  semGive(cids->tasks.<"$tname">.sync);
  if (sched)
    genom_<"$comp">_schedule_exec(cids);

  do {
    while(!cids->shutdown) {
      commonStructGive(cids);
<'if {[catch {$task period}]} {'>
      if (h2evnSusp(0) != TRUE) abort();
<'} else {'>
      if (h2timerPause(timer) != OK) abort();
<'}'>
      commonStructTake(cids);
      if (cids->tasks.<"$tname">.runnable) break;
    }
    cids->tasks.<"$tname">.runnable = 0;
    sched = 0;
    genom_log_debug("<"$tname"> tick");

    /* handle permanent activity */
    sched = genom_<"$comp">_<"$tname">_invoke_perm(cids);
    if (sched < 0) {
      cids->tasks.<"$tname">.status = cids->tasks.<"$tname">.pact;
      break;
    }

    /* handle current activities for this task */
    for(id=0; id<<"$COMP">_MAX_ACTIVITIES; id++) {
      a = &cids->activities.i[id].any;

      if (a->status == ACT_STOP) {
        a->status = ACT_STOPPING;
        a->codel = <"${comp}">_stop;
      }
      if (a->status != ACT_RUNNING && a->status != ACT_STOPPING)
        continue;

      a->codel = genom_<"$comp">_<"$tname">_invoke(cids, a);
      if (a->codel == NULL || a->codel == <"$comp">_ether) {
        if (a->status == ACT_RUNNING)
          a->status = ACT_ETHER;
        else
          a->status = ACT_STOPPED;
        if (a->codel == NULL)
          a->exdetail = (void *)genom_thrown(&a->codel);
        sched |= 1;
      } else if (a->codel != <"$comp">_sleep || a->status == ACT_STOP)
        sched |= 2;
    }

    if (sched & 1) genom_<"$comp">_schedule_cntrl(cids);
    if (sched & 2) genom_<"$comp">_schedule_exec(cids);
  } while(!cids->shutdown);

<'if {[llength [$task fsm stop]]} {'>
  /* stop permanent activity */
  if (cids->tasks.<"$tname">.pact !=
    <"[[[$component event] member ether] declarator]">) {
    cids->tasks.<"$tname">.pact =
      <"[[[$component event] member stop] declarator]">;
    do {
      if (genom_<"$comp">_<"$tname">_invoke_perm(cids)) {
        cids->tasks.<"$tname">.status = cids->tasks.<"$tname">.pact; break;
      }
    } while(cids->tasks.<"$tname">.pact !=
      <"[[[$component event] member ether] declarator]">);
  }
<'}'>

  genom_log_info("shutting down %s exec task", "<"$tname">");

  if (<"$comp">_<"$tname">_mbox_size()) {
    csMboxEnd();
  }
<'if {[catch {$task period} period]} {'>
  mboxEnd(0);
<'} else {'>
  h2timerStop(timer);
  h2timerFree(timer);
<'}'>
  cids->tasks.<"$tname">.taskid = ERROR;
  commonStructGive(cids);
  semGive(cids->tasks.<"$tname">.sync);
  return NULL;
}
